<?php

namespace Drupal\openstack_provider\Form\Config;

use Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface;
use Drupal\cloud\Service\CloudServiceInterface;
use Drupal\Component\Utility\Html;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Session\AccountInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * OpenStack launch form configurations.
 */
class OpenStackProviderAdminSettingsForm extends ConfigFormBase {

  /**
   * Drupal\Core\Config\ConfigFactoryInterface definition.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;

  /**
   * Drupal\cloud\Service\CloudServiceInterface definition.
   *
   * @var \Drupal\cloud\Service\CloudServiceInterface
   */
  protected $cloudService;

  /**
   * Drupal\Core\Messenger\MessengerInterface definition.
   *
   * @var \Drupal\Core\Messenger\MessengerInterface
   */
  protected $messenger;

  /**
   * The cloud service provider plugin manager (CloudConfigPluginManager).
   *
   * @var \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface
   */
  protected $cloudConfigPluginManager;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * Entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * OpenStackProviderAdminSettingsForm constructor.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   * @param \Drupal\cloud\Service\CloudServiceInterface $cloud_service
   *   The cloud service.
   * @param \Drupal\Core\Messenger\MessengerInterface $messenger
   *   The Messenger Service.
   * @param \Drupal\cloud\Plugin\cloud\config\CloudConfigPluginManagerInterface $cloud_config_plugin_manager
   *   The cloud service provider plugin manager (CloudConfigPluginManager).
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   * @param \Drupal\Core\Entity\EntityTypeManager $entity_type_manager
   *   The entity type manager object.
   */
  public function __construct(
    ConfigFactoryInterface $config_factory,
    CloudServiceInterface $cloud_service,
    MessengerInterface $messenger,
    CloudConfigPluginManagerInterface $cloud_config_plugin_manager,
    AccountInterface $current_user,
    EntityTypeManager $entity_type_manager
  ) {
    parent::__construct($config_factory);
    $this->configFactory = $config_factory;
    $this->cloudService = $cloud_service;
    $this->messenger = $messenger;
    $this->cloudConfigPluginManager = $cloud_config_plugin_manager;
    $this->currentUser = $current_user;
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('cloud'),
      $container->get('messenger'),
      $container->get('plugin.manager.cloud_config_plugin'),
      $container->get('current_user'),
      $container->get('entity_type.manager'),
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames(): array {
    return [
      'openstack_provider.settings',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'openstack_provider_admin_settings';
  }

  /**
   * Ajax callback function to update vpc, and subnets.
   *
   * This method is called when the target provider option changes.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function vpcCallback(array &$form, FormStateInterface $form_state) {
    return $form['aws']['network'];
  }

  /**
   * Ajax callback function to retrieve subnets.
   *
   * This method is called when the VPC option is changed.
   *
   * @param array $form
   *   An associative array containing the structure of the form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The current state of the form.
   */
  public function subnetCallback(array &$form, FormStateInterface $form_state) {
    return $form['aws']['network']['subnets'];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state): array {
    $config = $this->config('openstack_provider.settings');

    // Use the triggering_element to clear the #default_value of any #select
    // element that uses AJAX to update the select options.  Keep this
    // block of code in the buildForm() method.  If moved to the #ajax
    // callbacks, the form will sometimes be cached and cause an error.
    $triggering_element = $form_state->getTriggeringElement();
    $vpc_id = $config->get('vpc') ?? '';
    $target_provider = $form_state->getValue('target_provider') ?? $config->get('target_provider') ?? '';
    $public_subnet_1 = $form_state->getValue('public_subnet_1') ?? $config->get('public_subnet_1') ?? '';
    $private_subnet_1 = $form_state->getValue('private_subnet_1') ?? $config->get('private_subnet_1') ?? '';
    $private_subnet_2 = $form_state->getValue('private_subnet_1') ?? $config->get('private_subnet_2') ?? '';

    // Clear any selected values during an ajax callback.
    if (!empty($triggering_element)) {
      switch ($triggering_element['#name']) {
        case 'target_provider':
          $target_provider = $triggering_element['#value'];
          $vpc_id = '';
          $public_subnet_1 = '';
          $private_subnet_1 = '';
          $private_subnet_2 = '';
          break;

        case 'vpc':
          $vpc_id = $triggering_element['#value'];
          $public_subnet_1 = '';
          $private_subnet_1 = '';
          $private_subnet_2 = '';
          break;
      }
    }

    $form = parent::buildForm($form, $form_state);
    $form['openstack'] = [
      '#type' => 'details',
      '#title' => $this->t('OpenStack'),
      '#open' => TRUE,
    ];

    $providers = $this->getProviders('openstack');
    $form['openstack']['selectable_regions'] = [
      '#type' => 'select',
      '#title' => $this->t('Selectable OpenStack regions'),
      '#description' => $this->t('Regions available to users during tenant creation.'),
      '#options' => $providers,
      '#default_value' => $config->get('selectable_regions'),
      '#multiple' => TRUE,
      '#required' => TRUE,
      '#size' => count($providers),
    ];

    $form['aws'] = [
      '#type' => 'details',
      '#title' => $this->t('AWS Settings'),
      '#open' => TRUE,
    ];
    $form['aws']['target_provider'] = [
      '#type' => 'select',
      '#title' => $this->t('Target provider'),
      '#required' => TRUE,
      '#options' => $this->getProviders('aws_cloud', 'getCloudContext', TRUE),
      '#description' => $this->t('Select region to launch Cloud Orchestrator instances.'),
      '#ajax' => [
        'callback' => '::vpcCallback',
        'event' => 'change',
        'wrapper' => 'network-container',
      ],
      '#default_value' => $target_provider,
    ];

    $form['aws']['network'] = [
      '#type' => 'details',
      '#title' => $this->t('Network'),
      '#open' => TRUE,
      '#attributes' => [
        'id' => 'network-container',
      ],
    ];

    $form['aws']['network']['vpc'] = [
      '#type' => 'select',
      '#title' => $this->t('VPC'),
      '#description' => $this->t('Select a VPC to use with this stack.'),
      '#required' => TRUE,
      '#options' => $this->getVpcs($target_provider),
      '#default_value' => $vpc_id,
      '#ajax' => [
        'callback' => '::subnetCallback',
        'event' => 'change',
        'wrapper' => 'subnet-container',
      ],
    ];

    $form['aws']['network']['subnets'] = [
      '#type' => 'details',
      '#title' => $this->t('Subnets'),
      '#attributes' => [
        'id' => 'subnet-container',
      ],
      '#open' => TRUE,
    ];

    $form['aws']['network']['subnets']['public_subnet_1'] = [
      '#type' => 'select',
      '#title' => $this->t('Public subnet'),
      '#description' => $this->t('Select a public subnet to use with the EC2 instance.'),
      '#options' => $this->getSubnets($target_provider, $vpc_id),
      '#required' => TRUE,
      '#default_value' => $public_subnet_1,
    ];

    $form['aws']['network']['subnets']['private_subnet_1'] = [
      '#type' => 'select',
      '#title' => $this->t('Private subnet 1'),
      '#description' => $this->t('Select a private subnet for RDS, security groups and cache
      resources. Private Subnet 1 and Private Subnet 2 must be in two different
      Availability Zones in the same region.'),
      '#options' => $this->getSubnets($target_provider, $vpc_id),
      '#required' => TRUE,
      '#default_value' => $private_subnet_1,
    ];

    $form['aws']['network']['subnets']['private_subnet_2'] = [
      '#type' => 'select',
      '#title' => $this->t('Private subnet 2'),
      '#description' => $this->t('Select a private subnet for RDS, security groups and cache
      resources. Private Subnet 1 and Private Subnet 2 must be in two different
      Availability Zones in the same region.'),
      '#options' => $this->getSubnets($target_provider, $vpc_id),
      '#required' => TRUE,
      '#default_value' => $private_subnet_2,
    ];

    $form['cfn'] = [
      '#type' => 'details',
      '#title' => $this->t('CloudFormation'),
      '#open' => TRUE,
    ];

    $form['cfn']['instance_type'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Instance type'),
      '#description' => $this->t('The EC2 instance type'),
      '#default_value' => $config->get('instance_type'),
      '#required' => TRUE,
    ];

    $form['cfn']['cloud_orchestrator_version'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Cloud Orchestrator version'),
      '#description' => $this->t('Version of the Cloud Orchestrator to install.'),
      '#default_value' => $config->get('cloud_orchestrator_version'),
      '#required' => TRUE,
    ];

    $form['cfn']['creation_timeout'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Creation timeout'),
      '#description' => $this->t('The length of time that CloudFormation waits until template
      creation is considered unsuccessful and is deleted.'),
      '#default_value' => $config->get('creation_timeout'),
      '#required' => TRUE,
    ];

    $form['cfn']['docker_file_url'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Docker file url'),
      '#description' => $this->t('The Dockerfile used to build Cloud Orchestrator container'),
      '#default_value' => $config->get('docker_file_url'),
      '#required' => TRUE,
    ];

    $form['rds'] = [
      '#type' => 'details',
      '#title' => $this->t('RDS'),
      '#open' => TRUE,
    ];

    $form['rds']['rds_instance_type'] = [
      '#type' => 'textfield',
      '#title' => $this->t('RDS instance type'),
      '#default_value' => $config->get('rds_instance_type'),
      '#required' => TRUE,
    ];

    $form['rds']['rds_version'] = [
      '#type' => 'textfield',
      '#title' => $this->t('RDS MariaDB version'),
      '#default_value' => $config->get('rds_version'),
      '#required' => TRUE,
    ];
    $form['rds']['mysql_username'] = [
      '#type' => 'textfield',
      '#title' => $this->t('MySQL username'),
      '#default_value' => $config->get('mysql_username'),
      '#required' => TRUE,
    ];

    $form['rds']['database_name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Database name'),
      '#default_value' => $config->get('database_name'),
      '#required' => TRUE,
    ];

    $form['rds']['db_allocated_storage'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Database size'),
      '#description' => $this->t('The size of the database (Gb).  Must be between 5 and 16384.'),
      '#default_value' => $config->get('db_allocated_storage'),
      '#required' => TRUE,
    ];

    $form['drupal'] = [
      '#type' => 'details',
      '#title' => $this->t('Drupal'),
      '#open' => TRUE,
    ];

    $form['drupal']['drupal_site_name'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Site name'),
      '#default_value' => $config->get('drupal_site_name'),
      '#description' => $this->t('Site name for Cloud Orchestrator instance.'),
      '#required' => TRUE,
    ];

    $form['drupal']['drupal_site_email'] = [
      '#type' => 'email',
      '#title' => $this->t('Site email'),
      '#description' => $this->t('The From address in automated emails sent during registration and new password requests, and other notifications.'),
      '#default_value' => $config->get('drupal_site_email'),
      '#required' => TRUE,
    ];

    $zones = system_time_zones(NULL, TRUE);

    $form['drupal']['drupal_timezone'] = [
      '#type' => 'select',
      '#options' => $zones,
      '#title' => $this->t('Timezone'),
      '#default_value' => $config->get('drupal_timezone'),
      '#description' => $this->t('Default timezone.'),
      '#required' => TRUE,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $config = $this->configFactory()->getEditable('openstack_provider.settings');

    foreach ($form_state->cleanValues()->getValues() ?: [] as $key => $value) {
      if ($key === 'selectable_regions') {
        // This is a multiselect field.
        $config->set($key, $value);
      }
      else {
        $config->set($key, Html::escape($value));
      }
    }
    $config->save();
    parent::submitForm($form, $form_state);
  }

  /**
   * Get a list of cloud providers based on entity_bundle.
   *
   * @param string $entity_bundle
   *   The bundle to retrieve.
   * @param string $method_name
   *   The method name to act as the key.
   * @param bool $add_none
   *   Add an empty option.
   *
   * @return array
   *   An array list of providers.
   */
  private function getProviders(string $entity_bundle, string $method_name = 'id', bool $add_none = FALSE): array {
    $providers = [];
    if ($add_none === TRUE) {
      $providers['_none'] = $this->t('- Select -');
    }
    if (empty($entity_bundle)) {
      return [];
    }
    $entities = $this->cloudConfigPluginManager->loadConfigEntities($entity_bundle);
    foreach ($entities ?: [] as $entity) {
      if ($this->currentUser->hasPermission('view all cloud service providers') || $this->currentUser->hasPermission('view ' . $entity->getCloudContext())) {
        $providers[$entity->$method_name()] = $entity->label();
      }
    }
    return $providers;
  }

  /**
   * Get vpcs depending on target provider.
   *
   * @param string $target_provider
   *   The target provider.
   *
   * @return array
   *   An array of vpcs formatted for a select box.
   */
  private function getVpcs(string $target_provider): array {
    $vpcs = [
      '_none' => $this->t('- Select -'),
    ];
    try {
      $entities = $this->entityTypeManager
        ->getStorage('aws_cloud_vpc')
        ->loadByProperties(['cloud_context' => $target_provider]);

      foreach ($entities ?: [] as $entity) {
        if ($this->currentUser->hasPermission('view all cloud service providers') || $this->currentUser->hasPermission('view ' . $entity->getCloudContext())) {
          $vpcs[$entity->getVpcId()] = $this->t('%label', [
            '%label' => $entity->label(),
          ]);
        }
      }

    }
    catch (\Exception $e) {
      $this->cloudService->handleException($e);
      $vpcs = [];
    }
    return $vpcs;
  }

  /**
   * Get subnets depending on target provider.
   *
   * @param string $target_provider
   *   The target provider.
   * @param string $vpc_id
   *   The vpc_id to query from.
   *
   * @return array
   *   An array of subnets formatted for a select box.
   */
  private function getSubnets(string $target_provider, string $vpc_id): array {
    $subnets = [
      '_none' => $this->t('- Select -'),
    ];
    try {
      $entities = $this->entityTypeManager
        ->getStorage('aws_cloud_subnet')
        ->loadByProperties([
          'cloud_context' => $target_provider,
          'vpc_id' => $vpc_id,
        ]);

      foreach ($entities ?: [] as $entity) {
        if ($this->currentUser->hasPermission('view all cloud service providers') || $this->currentUser->hasPermission('view ' . $entity->getCloudContext())) {
          $subnets[$entity->getSubnetId()] = $this->t('%label - (%cidr)', [
            '%label' => $entity->label(),
            '%cidr' => $entity->getCidrBlock(),
          ]);
        }
      }
    }
    catch (\Exception $e) {
      $this->cloudService->handleException($e);
      $subnets = [];
    }
    return $subnets;
  }

}
